from typing import Tuple, List
import numpy as np
import math
from skimpy import FOV_Rectangle
from .load_specs import load_specs
from .detectors import DetectorSpecs
from .window_mapping import Coords, WindowMappingType
from .configuration_key import AltiusConfigurationKey


# ------------------------------------------------------------------------------
#           Class DetectorSpecs
# ------------------------------------------------------------------------------
class OpticsSpecs:
    """
    Stores optical properties about the ALTIUS instruments.  It internally stores an instance of the detector properties
    allowing it to map optical angles to pixels and vice-versa.
    """

    def __init__(self, config: AltiusConfigurationKey):

        channel = config.channel
        specs = load_specs('optics.yaml')
        yaml_specs = specs.subkey(channel)

        self._channel: str = channel
        self._detector_specs = DetectorSpecs(config)
        self._front_aperture_area: float = yaml_specs['front_aperture_area']
        self._wavelength_range_nm = yaml_specs.as_array('wavelength_range')

        fov_ll = Coords(y=yaml_specs['fov_size_y'][0], x=yaml_specs['fov_size_x'][0], windowtype=WindowMappingType.WINDOW_DETECTOR)
        fov_ur = Coords(y=yaml_specs['fov_size_y'][1], x=yaml_specs['fov_size_x'][1], windowtype=WindowMappingType.WINDOW_DETECTOR)
        noof_ll = Coords(y=yaml_specs['noof_size_y'][0], x=yaml_specs['noof_size_x'][0], windowtype=WindowMappingType.WINDOW_DETECTOR)
        noof_ur = Coords(y=yaml_specs['noof_size_y'][1], x=yaml_specs['noof_size_x'][1], windowtype=WindowMappingType.WINDOW_DETECTOR)

        self._fov_size = (fov_ll, fov_ur)
        self._noof_angular_size = (noof_ll, noof_ur)

        self._vignetting_profile_x = np.array([yaml_specs['vignetting_fov_x/values'],
                                               yaml_specs['vignetting_fov_x/angles']])
        self._vignetting_profile_y = np.array([yaml_specs['vignetting_fov_y/values'],
                                               yaml_specs['vignetting_fov_y/angles']])

    @property
    def channel(self):
        """
        The channel ('uv', 'vis' or 'nir') of this optics object
        """

        return self._channel

    @property
    def detector_specs(self):
        """
        The detector specifications of this optics channel
        """
        return self._detector_specs

    @property
    def front_aperture_area(self) -> float:
        """
        The area of the front aperture in cm2.
        """
        return self._front_aperture_area

    @property
    def fov_angular_size(self) -> Tuple[Coords, Coords]:
        """
        Specifies the square size of the in-field area of the optics. It is specified as a 2 element tuple. The 1st
        element specifies the (y,x) coordinates of the bottom left corner of the field of view measured in radians
        from the optic boresight. The 2nd element specifies the (y,x) coordinates of the top right corner of the field of
        view measured in radians from the optical boresight. Angles are positive upwards (and towards larger pixel numbers on the detector).

        Returns
        -------
        Tuple[Coords, Coords]
        """
        return self._fov_size

    @property
    def noof_angular_size(self) -> Tuple[Coords, Coords]:
        """
        Specifies the square size of the near out-of-field area of the optics. It is specified as a 2 element tuple.
        The 1st element specifies the (y,x) coordinates of the bottom left corner of the near out-of-field measured in radians
        from the optic boresight. The 2nd element specifies the (y,x) coordinates of the top right corner of the near out-of-field
        measured in radians from the optical boresight. Angles are positive upwards (and towards larger pixel numbers on the detector).

        Returns
        -------
        Tuple[ Coords, Coords]
        """
        return self._noof_angular_size

    @property
    def fullconv_angular_size(self) -> Tuple[Coords, Coords]:
        """
        The (Y,X) angular size of the full convolution window. It is a two element tuple. The first element
        provides the angles at the bottom left corner. The second element provides the angles at the
        upper right corner. All angles are in radians
        """

        ofsx = self._detector_specs.window_hierarchy.noof_window.Ofsx()
        ofsy = self._detector_specs.window_hierarchy.noof_window.Ofsy()
        nx = self._detector_specs.window_hierarchy.noof_window.Numx()
        ny = self._detector_specs.window_hierarchy.noof_window.Numy()
        fy = self._detector_specs.window_hierarchy.fullconv_window.Numy()
        fx = self._detector_specs.window_hierarchy.fullconv_window.Numx()

        y0 = np.array((self._noof_angular_size[0].y, self._noof_angular_size[0].x))     # Get bottom/left angle of noof region
        y1 = np.array((self._noof_angular_size[1].x, self._noof_angular_size[1].x))     # get top/right angle of noof region
        x0 = np.array([ofsy, ofsx])                                                     # FULLCONV coords of bottom left corner
        x1 = np.array([ofsy + ny, ofsx + nx])                                           # FULLCONV coords of top right
        xr = np.array([fy, fx])
        m = (y1 - y0) / (x1 - x0)
        c = y0 - m * x0                                                                 # value of angle when FULLCONV coord = (0,0)
        yr = m * xr + c                                                                 # value of angle at upper right of FULLCONV
        a0 = Coords(y=c[0], x=c[1], windowtype=WindowMappingType.RADIANS)
        a1 = Coords(y=yr[0], x=yr[1], windowtype=WindowMappingType.RADIANS)
        return (a0, a1)                                                            # return lower left angle and upper right angle

    @property
    def detector_pixels_per_radian(self):
        """
        The number of detector window pixels per radian. This angle is assumed to be identical
        for X and Y directions.
        """
        deltay = self._fov_size[1].y - self._fov_size[0].y
        return self._detector_specs.window_specs.Sy / deltay

    @property
    def radians_per_detector_pixel(self):
        """
        The number of radians per detector window pixel. This number is assumed to be identical for X and Y directions.
        :return:
        """
        return 1.0 / self.detector_pixels_per_radian

    # ------------------------------------------------------------------------------
    #           wavelength_range_nm
    # ------------------------------------------------------------------------------
    @property
    def wavelength_range_nm(self) -> np.ndarray:
        """
        The nominal wavelength range in nanometers of this instrument as a 2 element
        array. The first element is the lower wavelength range. The second element is the upper
        wavelength range.
        * [0] nominal start wavelength range in nm
        * [1] nominal end wavelength range.
        """
        return self._wavelength_range_nm

    # ------------------------------------------------------------------------------
    #           radians_per_pixelX
    # ------------------------------------------------------------------------------
    @property
    def radians_per_pixelX(self) -> float:
        """
        Returns the number of radians per pixel in the horizontal (X) directions as a float
        """
        return self.radians_per_detector_pixel

    # ------------------------------------------------------------------------------
    #           radians_per_pixelY
    # ------------------------------------------------------------------------------
    @property
    def radians_per_pixelY(self) -> float:
        """
        Returns the number of radians per pixel in the vertical (Y) direction as a float.
        """
        return self.radians_per_detector_pixel

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_delta_radians
    # ------------------------------------------------------------------------------
    def pixel_to_radians(self, pixel: Coords) -> Coords:
        """
        Converts a pixel location in any window to an angular offset in radians from the boresight.  All pixel locations
        must specify the centre of the pixel as occuring on the 0.5 value. Returns the angles as a Coords object.

        Parameters
        ----------
        pixel: Coords
            The x,y location in any of the windows defined in WindowMappingType. It does not support RADIANS or WINDOW_UNDEFINE

        Returns
        -------
        Coords
            The angles in radians of the pixel coordinates from the boresight.
        """

        detpixel = self._detector_specs.window_hierarchy.convert_to(pixel, WindowMappingType.WINDOW_DETECTOR, checkrange=False)
        boresight = self._detector_specs.boresight_location
        dthetax = (detpixel.x - boresight.x) * self.radians_per_pixelX
        dthetay = (detpixel.y - boresight.y) * self.radians_per_pixelY
        dtheta = Coords(x=dthetax, y=dthetay, windowtype=WindowMappingType.RADIANS)
        return dtheta

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_delta_radians
    # ------------------------------------------------------------------------------
    def radians_to_pixel(self, radians: Coords, destinationwindowtype: WindowMappingType) -> Coords:
        """
        Converts an angle from the boresight to a pixel coordinate in the requested window.

        Parameters
        ----------
        radians: Coords
            The x,y angle in radians. The angle in radians from the boresight. The windowtype of this Coords object should be RADIANS

        destinationwindowtype: WindowMappingType
            The window type of the returned coordinates. This should not be RADIANS or WINDOW_UNDEFINE

        Returns
        -------
        Coords
            The pixel cordinates of the angles w.r.t boresight in the destination window
        """
        boresight = self._detector_specs.boresight_location                                                             # Get the boresight in detector window corodinates
        detpixelx = radians.x / self.radians_per_pixelX + boresight.x                                                   # Get the X detector pixel location of this angle
        detpixely = radians.y / self.radians_per_pixelY + boresight.y                                                   # get the Y detector pixel location
        detcoord = Coords(x=detpixelx, y=detpixely, windowtype=WindowMappingType.WINDOW_DETECTOR)                       # Wrap this up as a detector coordinate
        destcoord = self._detector_specs.window_hierarchy.convert_to(detcoord, destinationwindowtype)                   # Transform to the target window
        return destcoord                                                                                                # and return.

    # ------------------------------------------------------------------------------
    #           make_column_of_pixel_fov
    # ------------------------------------------------------------------------------
    def make_column_of_pixel_fov(self,
                                 windowtype: WindowMappingType = WindowMappingType.WINDOW_SCIENCE,
                                 pixel_column: float = None,
                                 num_x_quadrature: int = 1,
                                 num_y_quadrature: int = 1) -> List[FOV_Rectangle]:
        """
        Calculates a list containing the FOV entries for pixels along a column of vertical pixels in the designated window.
        The list can be supplied to ``skimpy`` to create lines of sight for radiative transfer calculations.

        Parameters
        ----------
        windowtype: WindowMappingType
            Optional. The column of FOV pixels is made for the requested window. The default is WINDOW_SCIENCE. Note that the
            field of views are adjusted for the different sizes of pixels in difefrent windows. I.E. An L1C pixel is much wider (~3 x 150 for VIS)
            than a Science window pixel.
        pixel_column:  float
            Optional. If set then the column of pixels is taken at the given ``X`` location in the ``windowtype``. Note that the center of pixels
            occur on the 0.5 value. If not set (default) then the column is taken so it passes through the boresight location.
        num_x_quadrature: int
            Optional, default is 1. Specifies the number of quadrature points that should be used by skimpy when calculating radiance signals across the horizontal field of view.
        num_y_quadrature: int
            Optional, default is 1. Specifies the number of quadrature points that should be used by skimpy when calculating radiance signals across the vertical field of view.

        Returns
        -------
        List[FOV_Rectangle]
            A list of fov elements that can be used for skimpy model configuration. This list will be the same size as
            the vertical pixel dimension of the target window and maps one-to-one with those pixels.
        """

        usemap = self._detector_specs.window_hierarchy[windowtype]
        Ny = usemap.Numy()
        boreanglex = 0.0
        if (pixel_column is not None):
            userpt = Coords(x=pixel_column, y=Ny / 2, windowtype=windowtype)
            origin = self.pixel_to_radians(userpt)
            boreanglex = origin.x

        borerad = Coords(x=boreanglex, y=0.0, windowtype=WindowMappingType.RADIANS)
        origin = self.radians_to_pixel(borerad, windowtype)                                                     # Get the location o fthe boresight in this window.
        x0 = origin.x

        userpt = Coords(x=x0 - 0.5, y=origin.y, windowtype=windowtype)                                                  # Get the pixel location 0.5 pixels to the lefgt of the boresight
        leftangle = self.pixel_to_radians(userpt)                                                               # and get this location as an angle
        userpt = Coords(x=x0 + 0.5, y=origin.y, windowtype=windowtype)                                                  # get the pixel location 0.5 pixels to the right of the boresight
        rightangle = self.pixel_to_radians(userpt)                                                              # and get this location as an angle
        deltax = math.fabs(rightangle.x - leftangle.x)                                                                  # get the angular size of the X pixel

        fov = []
        userpt = Coords(x=x0, y=0.0, windowtype=windowtype)                                                             # start at the bottom of the detcetor
        bottomangle = self.pixel_to_radians(userpt)                                                             # Get the angle at the bottom edge of the first pixel
        for iy in range(Ny):                                                                                            # for each vertical pixel in the user window
            userpt = Coords(x=x0, y=(iy + 1), windowtype=windowtype)                                                    # get the user pixel coordinates at the top edge of the pixel
            topangle = self.pixel_to_radians(userpt)                                                            # get the angle at top edge of the pixel
            theta = (topangle.y + bottomangle.y) * 0.5                                                                  # Get the midpoint of the two angles. This is the boresight o fthe pixel
            deltay = (topangle.y - bottomangle.y)                                                                       # Get the angular range across the pixel. This is the delta angle
            entry = FOV_Rectangle(Hrange=(num_x_quadrature, 0.0, deltax), Vrange=(num_y_quadrature, theta, deltay))         # Create the FOV definition for the SKIMPY radiative transfer.
            fov.append(entry)                                                                                           # Append the FOV rectangle to our list
            bottomangle = topangle                                                                                      # The top of this pixel is the bottom of the next pixel.
        return fov
