from typing import Dict, List, Any, Union, Tuple
from datetime import datetime
import numpy as np
from skconfig import Configuration
import sktime
from .load_specs import load_specs
from .configuration_key import AltiusConfigurationKey


# ------------------------------------------------------------------------------
#           SRFUVSpecsDateEntry
# ------------------------------------------------------------------------------
class SRFUVSpecsDateEntry:
    """
    Stores the UV SRF entries for a given date. This is a list of fpi transmission curves.
    """
    def __init__(self, utc: datetime, yaml: Configuration):

        self._utc = utc
        self._description: str = yaml['description']
        self._fpi_transmission_filename: str = yaml['srf/fpi_transmission_curves/transmission_filename']
        self._tuneable_param_file: str = yaml['srf/fpi_transmission_curves/tuneable_param_file']
        self._fpi_airgaps: List[Tuple[float, float, float]] = yaml['srf/fpi_transmission_curves/airgaps']

    @property
    def utc(self) -> datetime:
        return self._utc

    @property
    def description(self) -> str:
        return self._fpi_transmission_filename

    @property
    def fpi_transmission_filename(self) -> str:
        return self._fpi_transmission_filename

    @property
    def airgaps(self) -> List[Tuple[float, float, float]]:
        return self._fpi_airgaps

    @property
    def tuneable_param_file(self) -> str:
        return self._tuneable_param_file


# ------------------------------------------------------------------------------
#           SRFUVSpecs
# ------------------------------------------------------------------------------
class SRFUVSpecs:
    """
    Stores the SRF specs for the UV channel.
    """

    def __init__(self, configkey: AltiusConfigurationKey):
        """

        Parameters
        ----------
        channel: str
            Specifies the Altius channel. It must be one of "vis", "uv", or "nir"

        date_value: str, datetime, np.datetime64 or float
            [Optional]. Specifies the UT date and time of the required PSF entry. The date and time must exactly match the
            date and time stored in the configuration file. It can be any representation of time that can be coerced by package ``sktime``
            into a valid datetime object. By default it is None and the last entry is loaded in.
        """

        self._date_entries: List[SRFUVSpecsDateEntry] = []                                        # The list of date entries
        self._current_date: SRFUVSpecsDateEntry = None                                            # The current active date entry
        config = load_specs('srf.yaml')

        yaml_specs = config.subkey('uv')
        for datesettings in yaml_specs.values:
            currentyaml = yaml_specs.subkey(datesettings)
            dateentry = SRFUVSpecsDateEntry(datesettings, currentyaml)
            self._date_entries.append(dateentry)
        if configkey.datestr is None:
            self.select_last_date()
        else:
            self.select_date(configkey.datestr)

    @property
    def current_entry(self) -> SRFUVSpecsDateEntry:
        return self._current_date

    # ------------------------------------------------------------------------------
    #           select_last_date
    # ------------------------------------------------------------------------------
    def select_last_date(self):
        self._current_date = self._date_entries[-1]

    # ------------------------------------------------------------------------------
    #           select_date
    # ------------------------------------------------------------------------------
    def select_date(self, datestr):

        utc = sktime.ut_to_datetime(datestr)
        ok = self._current_date.utc == utc if self._current_date is not None else False
        if not ok:
            index = None
            for i, value in enumerate(self._date_entries):
                if value.utc == utc:
                    index = i
                    break
            if index is not None:
                self._current_date = self._date_entries[index]
            else:
                raise "UTC entry {} was not found in the PSF configuration".format(datestr)
        return self._current_date

    # ------------------------------------------------------------------------------
    #           __getitem__
    # ------------------------------------------------------------------------------
    def __getitem__(self, utcval: Union[str, datetime]) -> SRFUVSpecsDateEntry:
        """
        Overloads the indexing operator. This can be used to change the current date of the settings if a date is given
        (str, datetime, np.datetime64, float over 2000) or it can be used to change the current wavelength in
        the current date settings (number given between 200 and 2000)

        Parameters
        -----------
        utcval: str, datetime, np.datetime64, str
            The UTC of the desired entry.

        Returns
        -------
        SRFUVSpecsDateEntry
            Returns the requested date entry.

        """
        utc = sktime.ut_to_datetime(utcval)
        self.select_date(utc)
        answer = self._current_date
        return answer
