import numpy as np
from .window_mapping import AltiusWindowSpecs, WindowMappingHierarchy, WindowMappingType, Coords


# ------------------------------------------------------------------------------
#           class WindowTransforms
# ------------------------------------------------------------------------------
class WindowTransforms:
    """
    Used to perform transform an image from one ALtius window to another. There are current 5 windows under consideration
    * NOOF window, the near out of field region specified at detector pixel resolution
    * Detector window, the full detector window
    * Science window, the infied region of the detector window. This is specifie din detector pixels
    * L1B window, this is a window that exactly overlaps the science window. There is binning in X and Y in science/detector pixels (typically 3)
    * L1C window. This is a window that exactly overlaps the science window. There is binning in X and Y in L1B pixels
    """

    def __init__(self, window_specs: AltiusWindowSpecs):
        self._hierarchy = WindowMappingHierarchy(window_specs)
        self._window_specs = window_specs

    # ------------------------------------------------------------------------------
    #           noof_to_science
    # ------------------------------------------------------------------------------
    def noof_to_science(self, noofimage: np.ndarray) -> np.ndarray:
        """
        Converts a near-out-of-field window image to a science window image

        Parameters
        ----------
        noofimage: np.ndarray
            The 2d near-out-of-field image

        Returns
        -------
        np.ndarray(Sy, Sx)
            Returns the corresponding science window image.
        """

        Ny, Nx = noofimage.shape
        assert (Ny == self._window_specs.NOOFy) and (Nx == self._window_specs.NOOFx)
        sci_to_noof = self._hierarchy.sci_window.CreateMappingToUpperLevel(self._hierarchy.noof_window)
        ofsy = sci_to_noof.Ofsy()
        ofsx = sci_to_noof.Ofsx()
        sy = sci_to_noof.Numy()
        sx = sci_to_noof.Numx()
        sci_image = noofimage[ofsy:ofsy + sy, ofsx:ofsx + sx]
        return sci_image

    # ------------------------------------------------------------------------------
    #           noof_to_science
    # ------------------------------------------------------------------------------
    def noof_to_detector(self, noofimage: np.ndarray) -> np.ndarray:
        """
        Converts a near-out-of-field window image to a detector window image

        Parameters
        ----------
        noofimage: np.ndarray
            The 2d near-out-of-field image

        Returns
        -------
        np.ndarray( Dy, Dx)
            Returns the corresponding detector window image.
        """
        Ny, Nx = noofimage.shape
        assert (Ny == self._window_specs.NOOFy) and (Nx == self._window_specs.NOOFx)
        detectorwindow = self._hierarchy.detector_window
        Dx = detectorwindow.Numx()
        Dy = detectorwindow.Numy()
        if (Dx > Nx):                                                                                                   # if the detector window is bigger than the near out of field window (UV)
            ofsx = -detectorwindow.Ofsx()
            ofsy = -detectorwindow.Ofsy()
            detectorimage = np.zeros([Dy, Dx])
            detectorimage[ofsy: ofsy + Ny, ofsx: ofsx + Nx] = noofimage
        else:
            ofsx = detectorwindow.Ofsx()
            ofsy = detectorwindow.Ofsy()
            detectorimage = noofimage[ofsy: ofsy + Dy, ofsx: ofsx + Dx]
        return detectorimage

    # ------------------------------------------------------------------------------
    #           fullconv_to_noof
    # ------------------------------------------------------------------------------
    def fullconv_to_noof(self, fullconvimage: np.ndarray) -> np.ndarray:
        """
        Converts a near-out-of-field window image to a detector window image

        Parameters
        ----------
        fullconvimage: np.ndarray
            The 2d fullconv image

        Returns
        -------
        np.ndarray( Dy, Dx)
            Returns the corresponding detector window image.
        """
        Fy, Fx = fullconvimage.shape
        assert (Fy == self._window_specs.FULLCONVy) and (Fx == self._window_specs.FULLCONVx)
        noofwindow = self._hierarchy.noof_window
        Nx = noofwindow.Numx()
        Ny = noofwindow.Numy()
        ofsx = noofwindow.Ofsx()
        ofsy = noofwindow.Ofsy()
        noofimage = fullconvimage[ofsy: ofsy + Ny, ofsx: ofsx + Nx]
        return noofimage

    # ------------------------------------------------------------------------------
    #           fullconv_to_noof
    # ------------------------------------------------------------------------------
    def noof_to_fullconv(self, noofimage: np.ndarray) -> np.ndarray:
        """
        Converts a near-out-of-field window image to a full convolution window.

        Parameters
        ----------
        fullconvimage: np.ndarray
            The 2d fullconv image

        Returns
        -------
        np.ndarray( Dy, Dx)
            Returns the corresponding detector window image.
        """
        Fx = self._hierarchy.fullconv_window.Numx()
        Fy = self._hierarchy.fullconv_window.Numy()
        fullconv = np.zeros([Fy, Fx])

        Ny, Nx = noofimage.shape
        assert (Ny == self._window_specs.NOOFy) and (Nx == self._window_specs.NOOFx)
        noofwindow = self._hierarchy.noof_window
        ofsx = noofwindow.Ofsx()
        ofsy = noofwindow.Ofsy()
        fullconv[ofsy: ofsy + Ny, ofsx: ofsx + Nx] = noofimage
        return fullconv

    # ------------------------------------------------------------------------------
    #           noof_to_l1c
    # ------------------------------------------------------------------------------
    def noof_to_l1c(self, noofimage: np.ndarray) -> np.ndarray:
        """
        Converts a near-out-of-field window image to an L1C image

        Parameters
        ----------
        noofimage: np.ndarray
            The 2d near-out-of-field image

        Returns
        -------
        np.ndarray(Ky, Kx)
            Returns the corresponding L1C image.
        """
        Ny, Nx = noofimage.shape
        assert (Ny == self._window_specs.NOOFy) and (Nx == self._window_specs.NOOFx)
        scisignal = self.noof_to_science(noofimage)
        l1csignal = self.science_to_l1c(scisignal)
        return l1csignal

    # ------------------------------------------------------------------------------
    #           fullconv_to_l1c
    # ------------------------------------------------------------------------------
    def fullconv_to_l1c(self, fullconvimage: np.ndarray) -> np.ndarray:
        """
        Converts a fullconv window image to an L1C image

        Parameters
        ----------
        fullconvimage: np.ndarray
            The 2d fullconv image

        Returns
        -------
        np.ndarray(Ky, Kx)
            Returns the corresponding L1C image.
        """
        Ny, Nx = fullconvimage.shape
        assert (Ny == self._window_specs.FULLCONVy) and (Nx == self._window_specs.FULLCONVx)
        noofimage = self.fullconv_to_noof(fullconvimage)
        scisignal = self.noof_to_science(noofimage)
        l1csignal = self.science_to_l1c(scisignal)
        return l1csignal

    # ------------------------------------------------------------------------------
    #           fullconv_to_detector
    # ------------------------------------------------------------------------------

    def fullconv_to_detector(self, fullconvimage: np.ndarray) -> np.ndarray:
        """
        Converts a full window image to a detector window image

        Parameters
        ----------
        detectorimage: np.ndarray
            The 2d detector image

        Returns
        -------
        np.ndarray( Ny, Nx)
            Returns the corresponding NOOF window image.
        """

        Fy, Fx = fullconvimage.shape
        assert (Fy == self._window_specs.FULLCONVy) and (Fx == self._window_specs.FULLCONVx)
        detectorwindow = self._hierarchy.detector_window.CreateMappingToUpperLevel(self._hierarchy.fullconv_window)
        fullconvwindow = self._hierarchy.fullconv_window
        Dx = detectorwindow.Numx()
        Dy = detectorwindow.Numy()

        ofsx = detectorwindow.Ofsx()
        ofsy = detectorwindow.Ofsy()

        detimage = np.array(fullconvimage[ofsy: ofsy + Dy, ofsx: ofsx + Dx])                            # then extract the detector window out of the full convolution window
        return detimage

    # ------------------------------------------------------------------------------
    #           detector_to_noof
    # ------------------------------------------------------------------------------
    def detector_to_fullconv(self, detectorimage: np.ndarray, centered_offsets: bool = False) -> np.ndarray:
        """
        Converts a near-out-of-field window image to a detector window image

        Parameters
        ----------
        detectorimage: np.ndarray
            The 2d detector image

        centered_offsets: bool
            If true then the detector image ignores the offset parameters in the window definition and is centers the image
            in the middle of the NOOF window. This is useful if making PSF images and you want it to be exactly centered.

        Returns
        -------
        np.ndarray( Ny, Nx)
            Returns the corresponding NOOF window image.
        """
        Dy, Dx = detectorimage.shape
        assert (Dy == self._window_specs.Dy) and (Dx == self._window_specs.Dx)
        detectorwindow = self._hierarchy.detector_window.CreateMappingToUpperLevel(self._hierarchy.fullconv_window)
        fullconvwindow = self._hierarchy.fullconv_window
        Fx = fullconvwindow.Numx()
        Fy = fullconvwindow.Numy()
        noofimage = np.zeros([Fy, Fx])
        if centered_offsets:
            ofsx = (Fx - Dx) // 2
            ofsy = (Fy - Dy) // 2
        else:
            ofsx = detectorwindow.Ofsx()
            ofsy = detectorwindow.Ofsy()

        if (Dx > Fx):                                                                                                   # if the detector window is bigger than the near out of field window (UV)
            noofimage[:, :] = detectorimage[-ofsy: -ofsy + Fy, -ofsx: -ofsx + Fx]                                        # then extract the fullconv window out of the detector window
        else:                                                                                                           # otherise detector window is smaller or ssame size than full conv
            noofimage[ofsy: ofsy + Dy, ofsx: ofsx + Dx] = detectorimage                                                 # so copy the entire detector window into the full conv window
        return noofimage

    # ------------------------------------------------------------------------------
    #           detector_to_l1c
    # ------------------------------------------------------------------------------
    def detector_to_science(self, detectorimage: np.ndarray) -> np.ndarray:
        """
        Converts a detector window image to a science window image

        Parameters
        ----------
        detectorimage: np.ndarray
            The 2d detector window image

        Returns
        -------
        np.ndarray( Sy, Sx)
            Returns the corresponding science window image.
        """

        Dy, Dx = detectorimage.shape
        assert (Dy == self._window_specs.Dy) and (Dx == self._window_specs.Dx)
        sci_to_detector = self._hierarchy.sci_window
        ofsx = sci_to_detector.Ofsx()
        ofsy = sci_to_detector.Ofsy()
        Sx = sci_to_detector.Numx()
        Sy = sci_to_detector.Numy()
        sci = detectorimage[ofsy: ofsy + Sy, ofsx: ofsx + Sx]
        return sci

    # ------------------------------------------------------------------------------
    #           detector_to_l1c
    # ------------------------------------------------------------------------------
    def detector_to_l1c(self, detectorimage: np.ndarray):
        """
        Converts a detector window image to an L1C image

        Parameters
        ----------
        detectorimage: np.ndarray
            The 2d detector window image

        Returns
        -------
        np.ndarray( Ky, Kx)
            Returns the corresponding L1C image.
        """

        Dy, Dx = detectorimage.shape
        assert (Dy == self._window_specs.Dy) and (Dx == self._window_specs.Dx)
        l1c_to_detector = self._hierarchy.l1c_window.CreateMappingToUpperLevel(self._hierarchy.detector_window)
        ofsx = l1c_to_detector.Ofsx()
        ofsy = l1c_to_detector.Ofsy()
        binx = l1c_to_detector.Binx()
        biny = l1c_to_detector.Biny()
        Kx = l1c_to_detector.Numx()
        Ky = l1c_to_detector.Numy()
        numx = Kx * binx
        numy = Ky * biny
        image = detectorimage[ofsy: ofsy + numy, ofsx: ofsx + numx].reshape([Ky, biny, Kx, binx])
        l1c = np.sum(image, axis=(1, 3))
        return l1c

    # ------------------------------------------------------------------------------
    #           science_to_l1c
    # ------------------------------------------------------------------------------
    def science_to_l1c(self, scienceimage: np.ndarray) -> np.ndarray:
        """
        Converts a science window image to an L1C image

        Parameters
        ----------
        scienceimage: np.ndarray
            The 2d science window image

        Returns
        -------
        np.ndarray( Ky, Kx)
            Returns the corresponding L1C image
        """

        Sy, Sx = scienceimage.shape
        assert (Sy == self._window_specs.Sy) and (Sx == self._window_specs.Sx)
        l1c_to_science = self._hierarchy.l1c_window.CreateMappingToUpperLevel(self._hierarchy.sci_window)
        binx = l1c_to_science.Binx()
        biny = l1c_to_science.Biny()
        Kx = l1c_to_science.Numx()
        Ky = l1c_to_science.Numy()
        image = scienceimage.reshape([Ky, biny, Kx, binx])
        l1c = np.sum(image, axis=(1, 3))
        return l1c

    # ------------------------------------------------------------------------------
    #           science_to_l1c
    # ------------------------------------------------------------------------------
    def l1c_to_science(self, l1c: np.ndarray) -> np.ndarray:
        """
        Converts a science window image to an L1C image

        Parameters
        ----------
        l1c: np.ndarray
            The 2d l1C window image

        Returns
        -------
        np.ndarray(Sy, Sx)
            Returns the corresponding science window image
        """

        Ky, Kx = l1c.shape
        assert (Ky == self._window_specs.Ky) and (Kx == self._window_specs.Kx)
        l1c_to_science = self._hierarchy.l1c_window.CreateMappingToUpperLevel(self._hierarchy.sci_window)
        binx = l1c_to_science.Binx()
        biny = l1c_to_science.Biny()
        Sx = self._hierarchy.sci_window.Numx()
        Sy = self._hierarchy.sci_window.Numy()
        image = np.zeros([Ky, biny, Kx, binx])
        for kx in range(Kx):
            for bx in range(binx):
                for by in range(biny):
                    image[:, by, kx, bx] = l1c[:, kx]
        science = image.reshape([Sy, Sx]) / (binx * biny)
        return science

    # ------------------------------------------------------------------------------
    #           science_to_l1c
    # ------------------------------------------------------------------------------
    def science_to_detector(self, science: np.ndarray) -> np.ndarray:
        """
        Converts a science window image to an detector image

        Parameters
        ----------
        science: np.ndarray
            The 2d science window image

        Returns
        -------
        np.ndarray( Dy, Dx)
            Returns the corresponding detector window image
        """

        Sy, Sx = science.shape
        assert (Sy == self._window_specs.Sy) and (Sx == self._window_specs.Sx)
        ofsx = self._hierarchy.sci_window.Ofsx()
        ofsy = self._hierarchy.sci_window.Ofsy()
        Dy = self._hierarchy.detector_window.Numy()
        Dx = self._hierarchy.detector_window.Numx()
        detector = np.zeros([Dy, Dx])
        detector[ofsy: ofsy + Sy, ofsx: ofsx + Sx] = science
        return detector

    # ------------------------------------------------------------------------------
    #           l1c_to_detector
    # ------------------------------------------------------------------------------
    def l1c_to_detector(self, l1c: np.ndarray) -> np.ndarray:
        """
        Converts an L1C image to a detector window image

        Parameters
        ----------
        l1c: np.ndarray
            The 2d l1C window image

        Returns
        -------
        np.ndarray( Dy, Dx)
            Returns the corresponding detector image
        """
        science = self.l1c_to_science(l1c)
        detector = self.science_to_detector(science)
        return detector
