
from typing import Union
from datetime import datetime
import os
import os.path
import numpy as np
import argparse
import sktime
from l1c_deconvolution_matrix.deconvolution_matrix import generate_convolution_matrix, generate_horizontal_stray_light_matrix, deconvolution_matrix_version_info
from l1c_deconvolution_matrix.matrix_file_io import write_to_netcdf, load_from_netcdf
from .. import __version__
from .. configuration import PSFSpecs, DetectorSpecs, OpticsSpecs, AltiusConfigurationKey, override_yaml_base_folder
from .. configuration.window_mapping import altius_window_specs_as_list
from . import PSFTools, ModifiedHarveyPSF


# ------------------------------------------------------------------------------
#           make_infield_convolution_matrix
# ------------------------------------------------------------------------------
def make_infield_convolution_matrix(channel: str,
                                    datevalue: str,
                                    wavelen: float,
                                    filename: str = None,
                                    version: str = None,
                                    overwrite: bool = False,
                                    debug_mode: bool = False):
    """
    This is a function that generates the L1C convolution matrix using the settings given in the *psf.yaml* and *detectors.yaml*
    configuration files. The code loads in all of the information from the configuration file, generates the convolution
    matrix by calling the *l1c_deconvolution_matrix* package and writes the resultant matrix to a netcdf file. BY default
    the code will not overwrite an existing file. The code performs most of the file i/o checks and permissions before calling
    the matrix generation so any access issues are identified early.

    The user must manually insert the name of the generated file into the *psf.yaml* settings file.
    """
    config = AltiusConfigurationKey(channel=channel, wavelength=wavelen, datestr=datevalue)
    utc = sktime.ut_to_datetime(config.datestr)
    wavelength_settings = PSFSpecs(config)[config.wavelength]                                                       # Fetch the PSF wavelength settings for the requested channel and date
    convolution_settings = wavelength_settings.convolution_matrix_settings                                              # Get the convolution settings from the wavelength entry
    detector_settings = DetectorSpecs(config, window_config_name=convolution_settings.detector_window_identification)   # Get the detector window setting using the settings described in the PSF settings
    array_window_specs_nt = detector_settings.window_specs                                                              # load in the window settings
    array_window_specs = altius_window_specs_as_list(array_window_specs_nt)                                             # and make it into a list we can pass to the convolution code.
    indices_filename = wavelength_settings.psf_interpolation_coeffs_filename.lower()

    Dy = array_window_specs_nt.Dy                                                                                       # get the size of the detector window
    Dx = array_window_specs_nt.Dx                                                                                       # both height and width in detector pixels (2000, 1504)
    cellpsf = []                                                                                                        # reset the number of PSF arrays used to make the convolution matrix
    locn = []                                                                                                           # reset the location of each of the PSF arrays on the detector window
    for index in range(wavelength_settings.num_images):                                                                       # for eac PSF image in teh configuration file
        psf_filespec = wavelength_settings[index]                                                                       # get the settings for this PSF file
        psffilename = psf_filespec.psf_filename                                                                         # get its filename
        location = psf_filespec.center_of_psf                                                                           # and its location on the detector window
        psf = PSFTools.load_psf_from_filename(psffilename)                                                              # load in the PSF
        assert (psf.shape[0] == Dy) and (psf.shape[1] == Dx)                                                            # make sure its the shape we expect.
        cellpsf.append(psf)                                                                                             # and append it to our list of PSF's
        locn.append([location.y, location.x, location.windowtype])                                                      # and append the location of the PSF

    location = np.array(locn)                                                                                           # convert the list of locations to an array (N, 3)
    if (indices_filename == 'undefined'):
        assert len(cellpsf) == 1, 'More than one PSF exists. You must use linear interpolation across the detector surface'
        coeffs = np.ones([1, Dy, Dx])                                               # Set all the interpolation coefficients to 1.0
        indices = np.zeros([1, Dy, Dx])                                             # Set all the PSF indices to zero.
    else:
        raise Exception("More than one PSF exists, len(cellpsf). This is not yet supported as the file I/O code for the indices and coefficients is not yet written")

    description = "{:s}_{:04d}-{:02d}-{:02d}_{:3d}".format(channel, utc.year, utc.month, utc.day, int(wavelen + 0.5))
    if (filename is None):
        basedir, name = os.path.split(psffilename)
        filename = os.path.join("{:s}".format(basedir), "psf-conv-matrix_{:s}.nc".format(description))

    filename = os.path.normpath(filename)
    if (not debug_mode):
        if os.path.exists(filename):
            if not overwrite:
                raise Exception("Cannot generate convolution matrix for {} as output file {} already exists. Edit configuration file <psf.yaml> or use optional parameter <overwrite>".format(description, filename))
            else:
                os.remove(filename)

    if version is None:
        version = "altius_l1pp_configuration Version [{}], l1c_deconvolution_matrix Version [{}]".format(__version__, deconvolution_matrix_version_info())

    if (debug_mode):
        print('Debug Mode: The destination filename is {}'.format(filename))
    else:
        a_matrix = generate_convolution_matrix(array_window_specs, cellpsf, location, coeffs, indices)
        print('Writing convolution matrix to file {}'.format(filename))
        write_to_netcdf(filename, channel, wavelen, a_matrix, array_window_specs, location, version=version, description=description)


# ------------------------------------------------------------------------------
#           make_infield_convolution_matrix
# ------------------------------------------------------------------------------

def make_horizontal_near_oof_matrix(channel: str,
                                    datevalue: str,
                                    wavelen: float,
                                    filename: str = None,
                                    version: str = None,
                                    overwrite: bool = False,
                                    debug_mode: bool = False):
    """
    This is an example that shows how to use the convolution code and compares the python implementation
    to the original matlab implementation to see if they give the same answer.
    """
    config = AltiusConfigurationKey(channel=channel, wavelength=wavelen, datestr=datevalue)
    utc = sktime.ut_to_datetime(config.datestr)
    psfspecs = PSFSpecs(config)[config.wavelength]                                                       # Fetch the PSF wavelength settings for the requested channel and date
    convolution_settings = psfspecs.convolution_matrix_settings                                              # Get the convolution settings from the wavelength entry
    detector_settings = DetectorSpecs(config, window_config_name=convolution_settings.detector_window_identification)  # Get the detector window setting using the settings described in the PSF settings
    optics = OpticsSpecs(config)
    array_window_specs_nt = detector_settings.window_specs                                                              # load in the window settings
    array_window_specs = altius_window_specs_as_list(array_window_specs_nt)                                             # and make it into a list we can pass to the convolution code.
    indices_filename = psfspecs.psf_interpolation_coeffs_filename.lower()
    vignetting_filename = psfspecs.psf_vignetting_array_filename

    Dy = array_window_specs_nt.Dy                                                                                       # get the size of the detector window
    Dx = array_window_specs_nt.Dx                                                                                       # both height and width in detector pixels (2000, 1504)
    cellpsf = []                                                                                                        # reset the number of PSF arrays used to make the convolution matrix
    locn = []                                                                                                           # reset the location of each of the PSF arrays on the detector window

    use_synthetic_psf = convolution_settings.horizontal_noof_matrix_settings['use_synthetic_harvey_psf']                # Should we use a synthetic PSF to model the near out of field. This
    if use_synthetic_psf:                                                                                               # should be done for vis and nir where the out-of-field region is much larger than the detcetor window
        harveycoeffs = psfspecs.far_field_harvey_coeffs
        psf, location = ModifiedHarveyPSF.fullconv_image_from_modified_harvey_curve(optics, harveycoeffs)
        cellpsf.append(psf)                                                                                             # and append it to our list of PSF's
        locn.append([location.y, location.x, location.windowtype])                                                    # and append the location of the PSF
        psffilename = psfspecs[0].psf_filename                                                                          # get the first psf filename so we have a default folder location
        indices_filename = 'undefined'                                                                                  # And make sure interpolation indices are not used.
    else:
        for index in range(psfspecs.num_images):                                                                              # for eac PSF image in teh configuration file
            psf_filespec = psfspecs[index]                                                                              # get the settings for this PSF file
            psffilename = psf_filespec.psf_filename                                                                     # get its filename
            location = psf_filespec.center_of_psf                                                                       # and its location on the detector window
            psf = PSFTools.load_psf_from_filename(psffilename)                                                          # load in the PSF
            assert (psf.shape[0] == Dy) and (psf.shape[1] == Dx)                                                        # make sure its the shape we expect.
            cellpsf.append(psf)                                                                                         # and append it to our list of PSF's
            locn.append([location.y, location.x, location.windowtype])                                                # and append the location of the PSF
    location = np.array(locn)                                                                                           # convert the list of locations to an array (N, 3)
    if (indices_filename == 'undefined'):
        Nx = array_window_specs_nt.NOOFx
        Ny = array_window_specs_nt.NOOFy
        assert len(cellpsf) == 1, 'More than one PSF exists. You must use linear interpolation across the detector surface'
        coeffs = np.ones([1, Ny, Nx])                                               # Set all the interpolation coefficients to 1.0
        indices = np.zeros([1, Ny, Nx])                                             # Set all the PSF indices to zero.
    else:
        raise Exception("More than one PSF exists. This is not yet supported as the file I/O code for the indices and coefficients is not yet written")

    if (vignetting_filename == 'undefined'):
        vignetting = np.zeros([0, 0])       # simulate an empty array. This is replaaced inside the code by 1.0 everywhere.
    else:
        raise Exception("Vignetting filename is specified This is not yet supported as the file I/O code for the vignetting arrays is not yet written")

    description = "{:s}_{:04d}-{:02d}-{:02d}_{:3d}".format(channel, utc.year, utc.month, utc.day, int(wavelen + 0.5))
    if (filename is None):
        basedir, name = os.path.split(psffilename)
        filename = os.path.join("{:s}".format(basedir), "psf-horiz-noof_{:s}.nc".format(description))

    filename = os.path.normpath(filename)
    if (not debug_mode):
        if os.path.exists(filename):
            if not overwrite:
                raise Exception("Cannot generate convolution matrix for {} as output file {} already exists. Edit configuration file <psf.yaml> or use optional parameter <overwrite>".format(description, filename))
            else:
                if debug_mode:
                    print('Existing file {} will be deleted and replaced in normal mode'.format(filename))
                else:
                    print('Deleting existing file {}'.format(filename))
                    os.remove(filename)

    if version is None:
        version = "altius_l1pp_configuration Version [{}], l1c_deconvolution_matrix Version [{}]".format(__version__, deconvolution_matrix_version_info())

    deltasteps = convolution_settings.horizontal_noof_matrix_settings
    if (debug_mode):
        print('Debug Mode: the destination filename is {}'.format(filename))
    else:
        a_matrix = generate_horizontal_stray_light_matrix(array_window_specs, vignetting, deltasteps, cellpsf, location, coeffs, indices)
        print('Writing horizontal noof matrix to file {}'.format(filename))
        write_to_netcdf(filename, channel, wavelen, a_matrix, array_window_specs, location, version=version, description=description)


# ------------------------------------------------------------------------------
#           make_convolution_matrix
# ------------------------------------------------------------------------------
def altius_l1pp_make_convolution_matrix(default_yaml_folder=""):

    """
    This is a function that provides a command line interface for generation of various L1C convolution matrices. The user
    must pre-configure the proper settings ``psf.yaml`` and ``detectors.yaml`` configuration files. The command line parameters
    guide how information is loaded from the configuration files and which type of convolution matrix is generated. By default,
    the code will not allow you to overwrite existing files and it performs most of the file i/o checks before calling
    the matrix generation code so access issues are identified prior to hours of computation.
    """

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--type', action='store', required=True, help='Select matrix type: convolution, horizontal or infield_stray')
    parser.add_argument('--channel', action='store', required=True, help='Select channel: uv, vis or nir')
    parser.add_argument('--wavelength', action='store', required=True, type=float, help='Select PSF wavelength. Must match an existing wavelength entry in the psf.yaml')
    parser.add_argument('--date', action='store', required=True, help='The data of the PSF entry, Typically in the format yyyy-mm-dd. Must match an existing entry in the psf.yaml')
    parser.add_argument('--filename', action='store', default=None, help='The output filename of the generated matrix. This will be a netcdf file.')
    parser.add_argument('--version', action='store', default=None, help='The version number. By default the version is extracted for the l1c_deconvolution_module')
    parser.add_argument('--overwrite', action='store_true', help='Allow existing files to be overwritten')
    parser.add_argument('--debug', action='store_true', help='Enable special debug mode.')
    parser.add_argument('--yaml_folder', action='store', help='The folder of the USASK YAML configuration files', default=default_yaml_folder)

    commandline = parser.parse_args()
    type = commandline.type.lower()
    channel = commandline.channel.lower()
    wavelen = commandline.wavelength
    datevalue = commandline.date
    filename = commandline.filename
    version = commandline.version
    overwrite = commandline.overwrite
    debug = commandline.debug

    yamlfolder = commandline.yaml_folder
    if (yamlfolder is not None) and (len(yamlfolder) > 0):
        override_yaml_base_folder(yamlfolder)

    if (type == 'convolution'):
        make_infield_convolution_matrix(channel, datevalue, wavelen, filename=filename, version=version, overwrite=overwrite, debug_mode=debug)
    elif (type == 'horizontal'):
        make_horizontal_near_oof_matrix(channel, datevalue, wavelen, filename=filename, version=version, overwrite=overwrite, debug_mode=debug)
    else:
        print('unrecognized <type> information [{}]'.format(type))
