from typing import Sequence, Tuple, Dict
import os.path
import xarray as xr
import numpy as np
from datetime import datetime
import scipy.io


# ------------------------------------------------------------------------------
#           _load_matlab_format
# ------------------------------------------------------------------------------
def _load_matlab_format(filename: str):
    psfs = scipy.io.loadmat(filename)
    psf = psfs.get('psf_uv')
    if psf is None:
        psf = psfs.get('psf_vis')
        if psf is None:
            psf = psfs.get('psf_nir')
            if psf is None:
                psf = psfs.get('PSF_perturbed')
    factor = np.sum(psf)
    psf /= factor
    return psf


# ------------------------------------------------------------------------------
#           write_to_netcdf
# ------------------------------------------------------------------------------
def _write_psf_to_netcdf(filename: str,
                         channel: str,
                         wavelength: float,
                         psf_arrays: np.ndarray,
                         psf_center_locn: np.ndarray,
                         utc: datetime = None,
                         version: str = "----",
                         description: str = "----"):
    """
    Writes the deconvolution matrix to a netcdf file. This should be compatible with legacy version created by earlier systems.

    Parameters
    ----------
    filename: str
        The full pathname of the netcdf file to be created. It will overwrite the file if it already exists

    channel: str
        A string identifying the channel. It will typically be one of 'uv', 'vis', 'nir'

    wavelength: float
        The wavelength in nanometers of the deconvolution matrix.

    psf_arrays: np.ndarray( Numpsf, Dy, Dx)
        An array of PSF arrays. This is the row major 2D square array generated by the deconvolution software. This array
        is written to the netcdf variable ``deconvolution_matrix``.

    psf_center_locn  : Array[Numpsf, 3].
        A 2-D array containing two numbers for each of the M measured PSF arrays.
        The three numbers store the center location of the corresponding PSF array and the type of coordinates, (Y,X, windowtype).
        ``windowtype`` values are, 2 corresponds to the NOOF window, 3 is the detcetor window and 3 is the science window. Note that
        0.0 corresponds to the left/bottom edge of the first pixel, 0.5 corresponds to the center of the first pixel and 1.0 is the
        right/upper edge of the first pixel. This array is written to the netcdf variable ``psf_location``.

    version: str
        An optional string providing version information. This value is written as as attribute of netcdf variable
        ``deconvolution_matrix``. The default value is "----"

    description: str
        An optional string providing description information. It is recommended that this field be used to describe
        whether the deconvolution matrix is infield, near-out-field etc. This value is written as an attribute of netcdf variable
        ``deconvolution_matrix`` The default value is "----"
    """
    if (utc is None):
        utc = datetime.utcnow().isoformat()                                                                                 # Get the time of file creation
    matrix_array = xr.DataArray(data=psf_arrays,
                                name='stray_psf',
                                dims=["NumPSF", "rows", "columns"],
                                attrs=dict(channel=channel,
                                           description=description,
                                           creation_time='{}'.format(utc),
                                           version=version,
                                           wavelength_nm=wavelength))

    psf_location = xr.DataArray(data=psf_center_locn,
                                name='psf_location',
                                dims=["NumPSF", "Coord"],
                                attrs=dict(description='Central Location of PSF'))

    coordnames = ['y', 'x', 'window_type']

    ds = xr.Dataset(data_vars=dict(psf=matrix_array, psf_location=psf_location), coords=dict(coordnames=(["Coord"], coordnames),))

    ds.to_netcdf(filename, encoding=dict(psf=dict(dtype='float32'),
                                         psf_location=dict(dtype='float64')))


# ------------------------------------------------------------------------------
#         load_from_netcdf
# ------------------------------------------------------------------------------
def _load_psf_from_netcdf(filename: str) -> Tuple[np.ndarray, Dict[str, int], np.ndarray]:
    """
    Loads the convolution matrix from a netcdf file. It is assumed this convolution matrix was written using
    method ``write_to_netcdf``

    Parameters
    ----------
    filename: str
        The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

    Returns
    -------
    Tuple[ convolution_matrix, window_specs, psf_location_array, approximation ]
        Returns a three element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
        as a numpy array. The 2nd element is a dictionary listing the window specifications used to create the convolution
        matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
        each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
    """

    ds = xr.open_dataset(filename)
    matrix = np.array(ds['deconvolution_matrix'])
    # approximation = None # int(ds['approximation'].data)
    locn = np.array(ds['psf_location'])
    specs = np.array(ds['window_specifications'])
    names = np.array(ds.coords['specsnames'])
    windowspecs = {}
    for i in range(names.size):
        windowspecs[names[i]] = int(specs[i])
    return matrix, windowspecs, locn
