

# ------------------------------------------------------------------------------
#           make_synthetic_psf
# ------------------------------------------------------------------------------
def make_synthetic_psf(y0: int, x0: int, Ny: int = 2000, Nx: int = 1504) -> np.ndarray:
    """
    Makes a point spread function that is quite similar to the PSF supplied by OIP in February 2020.
    Provides a convenient way to generate a point spread function without having to load in all of the
    SPS. The PSF is radially symmetric about the central point (y0,x0) and returns an image of
    size 2000,1504. The returned PSF is normalized so it sums to 1.000
    """

    #    Nx = 1504
    #    Ny = 2000
    r = np.arange(0, 752) + 0.5                                                  # Make an array of radius extending out to 752 pixels. The first element is at radius = 0.5 pixels.
    logpsf = np.zeros([752])                                                    # make an array to have the log of the PSF
    r0 = np.arange(0, 2)                                                        # The OIP PSF is represented by 4 regions
    r1 = np.arange(2, 6)                                                        # define the  pixel range of each region
    r2 = np.arange(6, 170)
    r3 = np.arange(170, 752)
    p0 = [-7.6064756095e+00, 2.8468404016e+00]                                  # Then get the polyval coefficients for each regions
    p1 = [-2.9924387649e-01, -8.1490048539e+00]                                 # These were taken from running altius_l1pp_mex_fit_synthetic_psf
    p2 = [2.3225352329e-04, -8.8649556867e-02, -9.5853915595e+00]
    p3 = [1.0584978215e-05, -1.5584362251e-02, -1.6016451104e+01]

    logpsf[r0] = np.polyval(p0, r[r0])                                          # Evaluate the poynomial coefficients
    logpsf[r1] = np.polyval(p1, r[r1])                                          # in each of the 4 regions
    logpsf[r2] = np.polyval(p2, r[r2])
    logpsf[r3] = np.polyval(p3, r[r3])

    x = np.arange(0, Nx).reshape([1, Nx])                                       # Now make a 2D array. Getthe values of X on a 0 based scale.
    y = np.arange(0, Ny).reshape([Ny, 1])                                       # Get the value sof Y on a 0 based scale
    xx = np.tile(x, [Ny, 1]) - x0                                               # get the value of DX for every location on a 2D array (Ny,Nx)
    yy = np.repeat(y, Nx, axis=1) - y0                                          # get the value of DY for every location on a 2D array (Ny,Nx)
    radius = np.sqrt(xx * xx + yy * yy)                                         # get the rdaial distance of every pixel from the center of the PSF
    psf = np.interp(radius, r, logpsf)                                          # Evaluate the log of PSF for every pixel. Out of bound points are set to NaN
    q = psf[(y0 - 4):(y0 + 4), (x0 - 4):(x0 + 4)]                               # Look for a Nan in the middle of the PSF, as the interpolation does not extrapolate below 0.5
    indices = np.where(np.isnan(q))                                             # if we have one
    for entry in indices:
        if entry.size > 0:
            row = entry[0]
            col = entry[1]
            iy = row + y0 - 5                                                   # get the index in to the main PSF array
            ix = col + x0 - 5                                                   # and set its
            psf[iy, ix] = logpsf[0]                                             # value  to the value at radius = 0.5
    bad = np.isnan(psf)                                                         # Now find all the other bad values
    psf[bad] = logpsf[751] - 1                                                  # and set them to a low, non-zero value .

    psf = np.exp(psf)                                                           # Convert log(stray_psf) back to stray_psf
    area = np.sum(psf)                                                          # Sum the area
    psf = psf / area                                                            # and normalize
    return psf
