from typing import Tuple, List
import numpy as np
import sasktran as sk
from skplatform import Platform
from skimpy import FrontEndOptics, SpectralWindow, WavelengthGaussianSRF, FrontEndRadianceGeneratorHR, RadianceValues
from ... configuration import OpticsSpecs, OrbitParams_SimpleOrbit, WindowMappingType, AltiusConfigurationKey


# ------------------------------------------------------------------------------
#           class AltiusL1ppSimpleSkimpyInterface
# ------------------------------------------------------------------------------
class AltiusL1ppSimpleSkimpyInterface:
    """
    A quick and simple interface of the ALTIUS instrument to the **skimpy** radiative transfer package.
    This class supports calculating vertical profiles of radiance for each of the 3 channels. Vertical profiles
    of radiance are calculated for each vertical pixel in the science window::

        >>> from altius_l1pp_preflight.configuration import AltiusConfigurationKey, OpticsSpecs
        >>> from altius_l1pp_preflight.radiance_simulator import AltiusL1ppSimpleSkimpyInterface
        >>>
        >>> configkey = AltiusConfigurationKey('vis')
        >>> optics = OpticsSpecs(configkey)
        >>> simulator = AltiusL1ppSimpleSkimpyInterface(optics, tp_latitude=38.0)
        >>> radiance, latitude, longitude, altitude = simulator.radiance(550.0, albedo=1.0)

    Users can optionally add custom settings for the platform during initalizion by providing any of the keywords used
    by :meth:`~.create_platform`. Note that all custom settings must be provided if this option is used.

    Parameters
    ----------
    optics: OpticsSpecs
        The optics specifications for the requested channel.
    """

    def __init__(self, optics: OpticsSpecs, **kwargs):
        self._optics: OpticsSpecs = optics
        self._orbitspecs = OrbitParams_SimpleOrbit()
        self._platform: Platform = None
        self._generator: FrontEndRadianceGeneratorHR = None
        self._atmosphere: sk.Atmosphere = None
        self._frontendoptics: FrontEndOptics = None
        self._radiance_values: List[RadianceValues] = None
        if (len(kwargs) > 0):
            self.create_platform(**kwargs)

    # ------------------------------------------------------------------------------
    #           make_altius_frontend
    # ------------------------------------------------------------------------------
    def _make_altius_frontendoptics(self, wavelen: float, fwhm: float = 1.3, num_wav_quadrature: int = 1) -> FrontEndOptics:

        front_end_optics = FrontEndOptics()
        wav0 = WavelengthGaussianSRF(num_wav_quadrature, wavelen, fwhm)                                                                    # wavelength region 1. 30 wavelengths across 450 nanometers, FWHM = 0.53 nm
        fovarray = self._optics.make_column_of_pixel_fov(WindowMappingType.WINDOW_SCIENCE)
        for fov in fovarray:
            front_end_optics.add_spectral_window(SpectralWindow(front_end_optics, fov, wav0))                              # define the spectral_windows used in the simulation
        return front_end_optics

    # ------------------------------------------------------------------------------
    #           make_atmosphere
    # ------------------------------------------------------------------------------
    def _default_altius_atmosphere(self, albedo: float) -> sk.Atmosphere:

        atmosphere = sk.Atmosphere()
        atmosphere.brdf = albedo
        atmosphere['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['ozone'] = sk.Species(sk.O3DBM(), sk.Labow())
        return atmosphere

    # ------------------------------------------------------------------------------
    #           make_altius_generator
    # ------------------------------------------------------------------------------
    def _default_altius_generator(self) -> FrontEndRadianceGeneratorHR:
        sasktran_options = {'top_of_atmosphere_altitude': 110000.0}
        generator = FrontEndRadianceGeneratorHR(sasktran_options=sasktran_options)
        return generator

    # ------------------------------------------------------------------------------
    #           def create_platform():
    # ------------------------------------------------------------------------------
    def create_platform(self, *,
                        utc: str = '2020-07-15 14:00:00.0000000',
                        tp_latitude: float = 40.0,
                        tp_longitude: float = 0.0,
                        tp_bearing_from_satellite: float = 187.0) -> Platform:
        """
        Creates the platform/satellite so it is looking at a desired tangent point at a given instant in time. This
        function can be called with no parameters and it will create mid-latitude, mid-afternoon, summertime conditions.
        The various parameters allow the satellite to be repositioned so it is looking at the target tangent point. Note
        that the position calculations are only nominal (as they are trick) and you may notice slight discrepancies in
        actual tangent point positions versus requested positiopn etc on the order of a few meters.

        Parameters
        ----------
        utc: str
            Set the universal time. This is typically used to position the Sun. The time is a string specifying UTC in iso-format.
            Default is '2020-07-15 14:00:00.0000000'.
        tp_latitude: float
            The latitude in degrees of the desired tangent point (**not** the satellite). Default is 40.0
        tp_longitude: float
            The longitude in degree o fthe desired tangent point. default is 0.0
        tp_bearing_from_satellite: float
            The geographic bearing (North = 0, East = 90) of the tangent point from the satellite. Default is 187.

        Returns
        -------
        Plaform
            The instance of class Platform created by the call. This is generally only required by advanced users. The
            value is also internally stored inside the class instance.
        """

        self._platform = Platform()
        tangent_latitude = tp_latitude
        tangent_longitude = tp_longitude
        geographic_bearing_of_tanpoint = tp_bearing_from_satellite
        tangent_height = self._orbitspecs.boresight_tangent_altitude
        satellite_height = self._orbitspecs.nominal_spacecraft_altitude
        assert (satellite_height < 800000.0)
        assert (satellite_height > 80000.0)

        utclist = [utc]
        observer = [tangent_latitude, tangent_longitude, tangent_height, geographic_bearing_of_tanpoint, satellite_height]
        tanpoint = [tangent_height, geographic_bearing_of_tanpoint]
        self._platform.add_measurement_set(utclist, ('looking_at_llh', observer), ('tangent_altitude', 'limb', tanpoint))
        return self._platform

    # ------------------------------------------------------------------------------
    #           radiance
    # ------------------------------------------------------------------------------
    def radiance(self, wavelength: float = None, *, albedo: float = 0.3, fwhm: float = 1.3) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculate the radiance for each pixel in the middle of the science window of the specified channel at the
        specified wavelength under the current observing conditions. Default values are used if no conditions have been previously set.
        The function also returns geo-location information for each of the pixels.

        Parameters
        ----------
        wavelength: float
            The wavelength in nm at which radiance should be calculated.
        albedo: float
            Optional. The albedo that should be used for the calculation, Default is 0.3
        fwhm: float
            Optional. The full-width Half-Max in nm of the gaussion SRF that should be convolved with thr solar signal. The default is 1.3 nm

        Returns
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            A four element tuple returning the radiance, and geo-location information. Each element contains a 1-D array which is the same size as the vertical size of the science
            window of the requested channel

            * [0] is the array of radiance
            * [1] is the latitude in degrees of the tangent point
            * [2] is the longitude in degrees of the tangent point
            * [3] is the altitude in meters of the tangent point
        """

        assert(wavelength is not None)

        self._atmosphere: sk.Atmosphere = self._default_altius_atmosphere(albedo)
        self._frontendoptics = self._make_altius_frontendoptics(wavelength, fwhm=fwhm)
        self._generator: FrontEndRadianceGeneratorHR = self._default_altius_generator()
        if (self._platform is None):
                self.create_platform()
        self._radiance_values = self._generator.calculate_radiance(self._frontendoptics, self._platform, self._atmosphere, toa_albedo_to_radiance_method='preconvolved_solar')

        geo = sk.Geodetic()
        npts = len(self._radiance_values)
        signal = np.zeros([npts])
        latitude = np.zeros([npts])
        longitude = np.zeros([npts])
        altitude = np.zeros([npts])

        for i, entry in enumerate(self._radiance_values):
            signal[i] = entry.radiance[0, 0, 0]
            pos = entry.ecef_position[:, 0, 0]
            look = entry.ecef_los[:, 0, 0]
            geo.from_tangent_point(observer=pos, look_vector=look)
            latitude[i] = geo.latitude
            longitude[i] = geo.longitude
            altitude[i] = geo.altitude
        return signal, latitude, longitude, altitude


if __name__ == '__main__':
    configkey = AltiusConfigurationKey('vis')
    optics = OpticsSpecs(configkey)
    simulator = AltiusL1ppSimpleSkimpyInterface(optics, tp_latitude=38.0)
    radiance, latitude, longitude, altitude = simulator.radiance(550.0, albedo=1.0)
    print('Done')
