from typing import Dict
import numpy as np
import sasktran as sk
from skimpy import Wavelength_GaussianPSF, Wavelength_Userdefined_SRF, SpectralWindow, Platform, RadianceValues_UVIS_NIR

from altius_l1pp_simulator.altius_bright_limb_signals.deprecated.skimpy_commonbase import SkimpyCommonBase
from  ..altius_srf.uv_srf import altius_uv_OOB_srf

#------------------------------------------------------------------------------
#           Skimpy_ALTIUSUV_OOB_StrayLight
#------------------------------------------------------------------------------

class SkimpyBrigh_tLimb(SkimpyCommonBase):
    """
    A class to implement radiative transfer calculations for a study of the ALTIUS UV spectral stray light issues.
    """

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self, mode: str ):
        super().__init__('uv', mode)
        self._wavelengths      : np.ndarray = None
        self._numwindows       : int        = None
        self._numheights       : int        = None

    #------------------------------------------------------------------------------
    #           calculate_signals
    #------------------------------------------------------------------------------

    def calculate_signals(self, albedo = 0.3, tanlatitude= 0.0, sza = 0.0, saa = 0.0, use_aerosols= True):

        self.make_instrument( )

        centralheight = 50000.0
        atmosphere    = self.make_atmosphere( albedo       = albedo,
                                              use_aerosols = use_aerosols )
        forwardmodel  = self.make_forwardmodel()
        platform,sun  = self.make_platform( centralheight = centralheight,
                                            tanlatitude   = tanlatitude,
                                            sza           = sza,
                                            saa           = saa  )

        forwardmodel_options={'solar_spectrum' : 'FONTELA_UVIS_3MICRON'}
        forwardmodel.calculate_radiance( self, platform, atmosphere, forward_model_options=forwardmodel_options, geometry_options={'sun':sun }  )

    #------------------------------------------------------------------------------
    #           _get_avg_radiance
    #------------------------------------------------------------------------------

    def _get_photon_flux(self, window:SpectralWindow)->float :

        hires_radiance = window.high_resolution_radiance
        avg_radiance   = RadianceValues_UVIS_NIR(   window, hires_radiance.radiance, hires_radiance.toa_albedo,
                                                    hires_radiance.wf,
                                                    False)
        return avg_radiance.photon_flux[0]

    #------------------------------------------------------------------------------
    #           num_windows
    #------------------------------------------------------------------------------
    @property
    def num_windows(self):
        return self._numwindows

    #------------------------------------------------------------------------------
    #           num_heights
    #------------------------------------------------------------------------------
    @property
    def num_heights(self):
        return self.spectral_windows.num_spectral_windows()//self._numwindows


    #------------------------------------------------------------------------------
    #           heights
    #------------------------------------------------------------------------------
    @property
    def heights(self):
        return self._backendsignals['height']


    #------------------------------------------------------------------------------
    #           get_spectral_window
    #------------------------------------------------------------------------------

    def get_spectral_window(self, height_index: int, window_index : int)->SpectralWindow :
        """
        Returns the spectral window at a given height index and window index. See the implementation
        of _make_instrument to get the specifics on the window index. The spectral window is useful to get access
        to the heigh resolution spectra. generated during teh RT calculations.

        :param height_index:
            The  height index of the required spectral window

        :param window_index:
             the index of the required window (typically 0-7 at time time of writing)
        :return:
        """

        window = self.spectral_windows[ height_index*self._numwindows + window_index]
        return window


    #------------------------------------------------------------------------------
    #           make_fpi_instrument
    #------------------------------------------------------------------------------

    def make_instrument( self ):

        """
        A function to make the spectral windows needed by skimpy to model ALTIUS UV in and out of band signals
        """

        self._numheights = self.make_height_profile_of_sciencepixel_fov(90.0)                                           # Make the field of view objects for a height profile covering 90 arc minutes.
        srf_wavelen, srf = altius_uv_OOB_srf( 320.0 )
        rt_wavelen       = np.arange(380.0, 1210.0, 10)                                                                 # These are the wavelengths specified for the radiative transfer calculation
                                                                                                                        # define the wavelength bandpasses for the spectral_windows. Simple spectrograph has several wavelength bandpasses
        #self.sbuv_startwinidx = self.add_wavelength_range( Wavelength_GaussianPSF(1, 251.9, 3.0))                       # add the wavelength range object for SBUV wavelength 1
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 273.5, 3.0))                                               # add the wavelength range object for SBUV wavelength 2
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 283.0, 3.0))                                               # add the wavelength range object for SBUV wavelength 3
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 287.6, 3.0))                                               # add the wavelength range object for SBUV wavelength 4
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 292.2, 3.0))                                               # add the wavelength range object for SBUV wavelength 5
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 297.5, 3.0))                                               # add the wavelength range object for SBUV wavelength 6
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 301.9, 3.0))                                               # add the wavelength range object for SBUV wavelength 7
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 305.8, 3.0))                                               # add the wavelength range object for SBUV wavelength 8
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 312.5, 3.0))                                               # add the wavelength range object for SBUV wavelength 9
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 317.5, 3.0))                                               # add the wavelength range object for SBUV wavelength 10
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 331.2, 3.0))                                               # add the wavelength range object for SBUV wavelength 11
        #self.add_wavelength_range( Wavelength_GaussianPSF(1, 339.8, 3.0))                                               # add the wavelength range object for SBUV wavelength 12


        self.inband_startwinidx = self.add_wavelength_range( Wavelength_GaussianPSF(1, 250.0, 3.0,
                                                                                    photon_distribution='additive'))    # add the wavelength range object for ALTIUS UV wavelength 1
        self.add_wavelength_range( Wavelength_GaussianPSF(1, 300.0, 3.0, photon_distribution='additive'))               # add the wavelength range object for ALTIUS UV wavelength 2
        self.add_wavelength_range( Wavelength_GaussianPSF(1, 315.0, 3.0, photon_distribution='additive'))               # add the wavelength range object for ALTIUS UV wavelength 3
        self.add_wavelength_range( Wavelength_GaussianPSF(1, 351.0, 3.0, photon_distribution='additive'))               # add the wavelength range object for ALTIUS UV wavelength 4

        self.outofband_winidx =  self.add_wavelength_range( Wavelength_Userdefined_SRF( srf, srf_wavelen, 850,          # add the wavelength range object for the ALTIUS UV out-of-band spectral stray light (mostly from the NIR)
                                                                                        rt_wavelen = rt_wavelen,
                                                                                        rtm_processing_scheme='hybrid',
                                                                                        photon_distribution='additive'))
        self.proxy_winidx     = self.add_wavelength_range( Wavelength_GaussianPSF( 1,  675.0, 3.0,
                                                                                   photon_distribution='additive'))     # add the wavelength range object for the NIR proxy signal

        for i in range(self._numheights):                                                                               # for each height
            #self.add_spectral_window( SpectralWindow(self,  i,  0, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  1, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  1, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  2, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  2, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  3, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  3, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  4, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  4, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  5, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  5, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  6, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  6, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  7, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  7, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  8, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  8, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength  9, at this height
            #self.add_spectral_window( SpectralWindow(self,  i,  9, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength 10, at this height
            #self.add_spectral_window( SpectralWindow(self,  i, 10, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength 11, at this height
            #self.add_spectral_window( SpectralWindow(self,  i, 11, self.aperture_area_cm2))                             # create the spectral window for SBUV wavelength 12, at this height

            self.add_spectral_window( SpectralWindow(self,  i, 0, self.aperture_area_cm2))                             # create the in-band spectral window for ALTIUS UV Wavelength 1 at this height
            self.add_spectral_window( SpectralWindow(self,  i, 1, self.aperture_area_cm2))                             # create the in-band spectral window for ALTIUS UV Wavelength 2 at this height
            self.add_spectral_window( SpectralWindow(self,  i, 2, self.aperture_area_cm2))                             # create the in-band spectral window for ALTIUS UV Wavelength 3 at this height
            self.add_spectral_window( SpectralWindow(self,  i, 3, self.aperture_area_cm2))                             # create the in-band spectral window for ALTIUS UV Wavelength 4 at this height

            self.add_spectral_window( SpectralWindow(self,  i, 4, self.aperture_area_cm2))                             # create the wide out-of-bad spectral window for ALTIUS UV at this height.
            self.add_spectral_window( SpectralWindow(self,  i, 5, self.aperture_area_cm2))                             # create the proxy spectral window for ALTIUS UV at this height.
        return self

    # ------------------------------------------------------------------------------
    #           propagate_frontend_to_backend
    # ------------------------------------------------------------------------------

    def propagate_frontend_to_backend(self, platform: Platform, attitude_solution: Dict[str, np.ndarray]):
        """
        The callback function used by skimpy to allow the instrument_specs to organize the high resolution
        radiances calculed by the forward model.
        """

        self._numwindows = self.wavelength_ranges.size
        self._numheights = self.spectral_windows.num_spectral_windows() // self._numwindows

        numheights = self._numheights
        numwindows = self._numwindows
        numib      = self.num_inband_windows
        photons_IB = np.zeros([numib, numheights])
        photons_OOB = np.zeros([numheights])
        photons_PXY = np.zeros([numheights])
        wavelen_ib   = np.zeros([numib])
        latitude = np.zeros([numheights])
        longitude = np.zeros([numheights])
        height = np.zeros([numheights])
        geo = sk.Geodetic()

        for ih in range(numheights):

            windowOOB = self.spectral_windows[ih * numwindows + self.outofband_winidx]  # Get the Out-of-band spectral window
            windowPXY = self.spectral_windows[ih * numwindows + self.proxy_winidx]  # Get the proxy spectral window
            photons_OOB[ih] = self._get_photon_flux(windowOOB)
            photons_PXY[ih] = self._get_photon_flux(windowPXY)


            for ibidx in range( numib):
                windowIB = self.spectral_windows[ ih * numwindows + self.inband_startwinidx + ibidx]  # Get the In-Band spectral window
                photons_IB[ibidx, ih] = self._get_photon_flux(windowIB)
                wavelen_ib[ibidx]      = windowIB.central_wavelength()

            look = attitude_solution['los'][:, ih, 0]
            pos = attitude_solution['pos'][:, ih, 0]
            geo.from_tangent_point(pos, look)
            latitude[ih] = geo.latitude
            longitude[ih] = geo.longitude
            height[ih] = geo.altitude

        utc = attitude_solution['utc'][0, 0]
        self._backendsignals['photons_IB'] = photons_IB
        self._backendsignals['photons_OOB'] = photons_OOB
        self._backendsignals['photons_PXY'] = photons_PXY
        self._backendsignals['wavelength_IB'] = wavelen_ib
        self._backendsignals['wavelength_OOB'] = windowOOB.wavelength_range()
        self._backendsignals['wavelength_PXY'] = windowPXY.central_wavelength()
        self._backendsignals['latitude'] = latitude
        self._backendsignals['longitude'] = longitude
        self._backendsignals['height'] = height
        self._backendsignals['utc'] = utc


