from typing import Union, Tuple
import os.path
from collections import namedtuple
import xarray as xs
import numpy as np
from ... configuration import OpticsSpecs, WindowTransforms, AltiusConfigurationKey


_StrayLightCaseNames = namedtuple('StrayLightCaseNames', ['name', 'sza', 'raa'])


# ------------------------------------------------------------------------------
#           SimulationRadianceDataBISA
# ------------------------------------------------------------------------------
class SimulationRadianceDataBISA:
    """
    This is a class used to manage the radiances used for the stray light sensitivity analysis. It loads in the
    the stimuli generated by BISA (Emmanuel Dekemper, February 2021). Email message from Emmanuel on 2021-02-15 message
    is below.

    Please find attached the radiances computed as agreed for this sensitivity analysis (thanks to my colleague
    Philippe for generating them!). There are 16 files stored in the archive: 8 positions along the bright limb
    side of the orbit times 2 for the 2 ozone profiles. Albedo = 1 was assumed everywhere.

    By opening the archive, you will find the following things,

    * A text document listing the full name of the 16 stimuli, and the average solar angles that produced the
      radiances (because they are not written in the file name, though they provide meaningful contextual information).
    * The 16 stimuli files. They are netcdf files, which you can most easily open in Matlab by using the method
      'bira.sim.scenes.LimbStimuli.import' (in the SPS package). This should provide you with a stimuli object which
      is pretty much like a Matlab structure. Of most interest to you: the 'radiance' field which has 4x6xN dimensions.
      Only the data found in radiance(1,:,:) are relevant for you (the rest is just there to mimick a Stokes vector).
      The 2nd dimension is for the 6 wavelengths (see 'wlGrid' field). The last dimension allows for the dependence
      of the radiance to the zenith angle of the line of sight (see 'losza' field).

    """

    def __init__(self, case: int = None):
        self.case_names: _StrayLightCaseNames = self._load_headers()
        self._angle: np.ndarray = None
        self._altitude: np.ndarray = None
        self._radiance: np.ndarray = None
        self._wavelen: np.ndarray = None
        self._stimulus_index: int = -1
        if (case is not None):
            self.load_stimulus(case)

    @property
    def num_cases(self):
        return self.case_names.sza.size

    @property
    def radiance(self):
        """
        The radiance profile for the heights and wavelengths contained in the stimulus. This is a 2-D array (Nh, Nw)
        where Nh is the number of heights (lines of sights) and Nw is the number of wavelengths
        """
        return self._radiance

    @property
    def angle(self):
        """
        A 1-D array(Nh) of angles. Each element is the angle in radians of each line of sight from the central boresight (looking
        at 50 km). It is the same size as the altitudes.
        """
        return self._angle

    @property
    def wavelength(self):
        """
        A 1-D array(Nw) of wavelengths in nanometers.
        """
        return self._wavelen

    @property
    def altitude(self):
        """
        A 1-D array(Nh) of nominal tangent altitudes in kilometers.
        """
        return self._altitude

    # ------------------------------------------------------------------------------
    #           _load_headers
    # ------------------------------------------------------------------------------
    def _load_headers(self) -> _StrayLightCaseNames:
        basedir, name = os.path.split(__file__)
        filename = os.path.join(basedir, 'RadiancesForNOOFStraylight', 'StimFilesProperties.dat')
        textdata = np.loadtxt(filename, dtype=str, delimiter=',')
        names = np.array(textdata[1:, 0])
        sza = textdata[1:, 1].astype(float)
        raa = textdata[1:, 2].astype(float)
        cases = _StrayLightCaseNames(name=names, sza=sza, raa=raa)
        return cases

    # ------------------------------------------------------------------------------
    #           angle_to_altitude
    # ------------------------------------------------------------------------------
    def angle_to_altitude(self, angles) -> np.ndarray:
        """
        Converts users supplied angles to altitude. It interpolates the values provided in the stim,ulus.
        """
        return np.interp(angles, self._angle, self._altitude, left=np.nan, right=np.nan)

    # ------------------------------------------------------------------------------
    #           load_stimulus
    # ------------------------------------------------------------------------------
    def load_stimulus(self, filename_or_case_number: Union[str, int]):
        """
        Loads in one of the stray light stimulus radiances used to test stray light
        removal. You can either specify the full filename o fthe stimulus or you can specify its index number (0-15)
        (ie its line number in the file *stimfilesproperties.dat*.

        The method loads in the new set of radiances, angles, altitudes and wavelengths from the case file

        Parameters
        ----------
        filename_or_case_number: int or str
            Either an integer specifying the case number (0-15) or a string representing the fullname of the file.
        """

        if (type(filename_or_case_number) is int):
            index = filename_or_case_number
            already_loaded = self._stimulus_index == index
            filename = self.case_names.name[filename_or_case_number]
            basedir, name = os.path.split(__file__)
            filename = os.path.join(basedir, 'RadiancesForNOOFStraylight', filename)
        else:
            index = -1
            filename = filename_or_case_number
            already_loaded = False

        if (not already_loaded):
            self._stimulus_index = index
            dataset = xs.open_dataset(filename)
            self._altitude = np.array(dataset['apparent_tp_altitude'])
            za = np.array(dataset['apparent_los_za'])
            self._radiance = np.array(dataset['radiance'][:, :, 0])
            self._wavelen = np.array(dataset['wavelength'])
            h50 = np.abs(50.0 - self._altitude)
            boreindex = np.argmin(h50)
            self._angle = np.radians(za[boreindex] - za)

            self._angle = np.flip(self._angle)
            self._altitude = np.flip(self._altitude)
            self._radiance = np.array(np.flip(self._radiance, axis=0), dtype='float64')

    # ------------------------------------------------------------------------------
    #           make_noof_image
    # ------------------------------------------------------------------------------
    def make_noof_image(self, opticsspecs: OpticsSpecs, stimulus_index: int = 4, wavelength_index: int = 4, maxvalue=30000.0) -> Tuple[np.ndarray, np.ndarray]:
        """
        Makes an image the size of the Near-out-of-field regioon and fills it with the radiance data for the
        given case number and wavelength index.
        """

        noof_size = opticsspecs.noof_angular_size
        start_noof = noof_size[0].y
        end_noof = noof_size[1].y
        window_specs = opticsspecs.detector_specs.window_specs
        self.load_stimulus(stimulus_index)
        angle = self.angle
        radiance = self.radiance[:, wavelength_index]

        fixed_angle = np.linspace(start_noof, end_noof, window_specs.NOOFy)
        signal = np.interp(fixed_angle, angle, radiance)
        signal *= maxvalue / np.max(signal)

        NOOFx = window_specs.NOOFx
        NOOFy = window_specs.NOOFy
        image = np.zeros([NOOFy, NOOFx])                                                            # Then create an array the size of the near out of field
        for i in range(NOOFx):                                                                      # and insert the 1-D profile
            image[:, i] = signal                                                                    # into every column
        return image, fixed_angle

    # ------------------------------------------------------------------------------
    #           make_noof_image
    # ------------------------------------------------------------------------------
    def make_fullconv_image(self, config: AltiusConfigurationKey, stimulus_index: int = 4, wavelength_index: int = 4, maxvalue=30000.0) -> Tuple[np.ndarray, np.ndarray]:
        """
        Makes an image the size of the Near-out-of-field regioon and fills it with the radiance data for the
        given case number and wavelength index.
        """

        opticsspecs = OpticsSpecs(config)
        noofimage, noofangles = self.make_noof_image(opticsspecs, stimulus_index=stimulus_index, wavelength_index=wavelength_index, maxvalue=maxvalue)
        fullconvimage = opticsspecs.detector_specs.window_transforms.noof_to_fullconv(noofimage)

        angular_size = opticsspecs.fullconv_angular_size                                                                # Get the angular size of the full convolution window
        start_angle = angular_size[0].y                                                                                 # Get the start angle
        end_angle = angular_size[1].y                                                                                 # Gte the end angle
        Fy = opticsspecs.detector_specs.window_specs.FULLCONVy
        fixed_angle = np.linspace(start_angle, end_angle, Fy)
        return fullconvimage, fixed_angle


if __name__ == "__main__":
    stimuli = SimulationRadianceDataBISA()
    stimuli.load_stimulus(3)
    print('Done')
