from typing import Union, Tuple
import numpy as np
import xarray as xr


class SrfInterface:
    """
    Generic class defining main functions for handling common operations
    on the SRF.  Due to the difference in the instruments used for the
    UV and VIS/NIR, this interface was deemed necessary as the implemenation
    of the SRF is quite different between the bands.  Where the UV has very
    non-Gaussian point spread functions at each wavelength, the VIS/NIR
    essentially have Gaussians as point spread functions.

    This is needed for simulating the scenes where a convolution with the full
    SRF is neeeded, and that is about it (as we don't do any corrections to the
    raw simulation)
    """

    LINEAR_INTERP = 'linear'
    LOG_INTERP = 'log'

    def get_srf(self, center_wavelength_nm, new_wavelengths=None,
                interp_method=LINEAR_INTERP, normalize=True) -> Tuple[xr.DataArray, float]:
        """
        Returns the SRF closest to the specified wavelength, and the actual wavelength

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength to obtain the SRF from the tunable wavelengths

        new_wavelengths : np.ndarray, Optional
            The grid of wavelengths, in nm, to evaluate/interpolate the SRF over.  If not
            specified, will return the SRF at the original measured wavelengths in wavelength_grid

        interp_method : str, Optional
            'linear' or 'log'.  Interpolation method for evaluating the SRF over new_wavelengths

        normalize : bool, Optional
            If True, normalizes the integral of the SRF to be 1.0.
            If False, the values of the SRF are adjusted so that the integral matches the original
            data's integral

        Returns
        -------
        srf : xr.DataArray
            The SRF with tuned wavelength near to center_wavelength_nm.  x-coordinate 'wavelength' set
            by the new_wavelengths parameter
        actual_wavelength : float
            The nearest tunable wavelength to the input center_wavelength_nm
        """
        pass

    def convolve_spectrum_with_srf(self, convolution_wavelengths: Union[float, np.ndarray], spectrum_wavelengths, spectrum) -> np.ndarray:
        """
            Convolves the spectrum with the SRF at each wavelength.

            Parameters
            ----------
            spectrum_wavelengths : np.ndarray
                Wavelengths in nm of the input spectrum

            spectrum : np.ndarray
                Input high-resolution spectrum to be convolved with the corrected SRF

            convolution_wavelengths : Union[float, np.ndarray]
                Wavelengths (in nm) to evaluate the resulting convolution at

            Returns
            -------
            convolved_spectrum : np.ndarray
                The resulting spectrum convolved with SRFs at each convolution wavelength
        """
        pass

    def get_wavelength_grid(self, resolution=None):
        pass