from typing import Tuple, List
import os.path
import numpy as np
import pickle
import scipy.io
import scipy.signal
import scipy.interpolate
import matplotlib.pyplot as plt
import sktime
from altius_l1pp_configuration.psf import PSFTools, L1CConvolutionMatrix
from altius_l1pp_configuration.configuration import PSFSpecs, DetectorSpecs

# ------------------------------------------------------------------------------
#           make_l1c_from_detector_window_image
# ------------------------------------------------------------------------------
def make_l1c_from_detector_window_image(image, array_window_specs):
    """
    MAkes an L1C product given a detector window and a set of specs.
    """

    SOFSx = array_window_specs[6]
    SOFSy = array_window_specs[7]
    Sx = array_window_specs[8]
    Sy = array_window_specs[9]
    LBINx = array_window_specs[10]
    LBINy = array_window_specs[11]
    KBINx = array_window_specs[14]
    KBINy = array_window_specs[15]
    Kx = array_window_specs[16]
    Ky = array_window_specs[17]
    newimage = image[SOFSy:SOFSy + Sy, SOFSx:SOFSx + Sx]                    # Make a new image that collects just the science window from the detector window
    l1b = newimage.reshape([Ky, LBINy * KBINy, Kx, LBINx * KBINx])          # Reshape Y axis into two components and X axis into two components
    l1c = np.sum(l1b, axis=(1, 3))                                          # Sum over the axes to leave the L1C product.
    return l1c                                                              # and return the L1C product


# ------------------------------------------------------------------------------
#           L1C_demo_UV
# ------------------------------------------------------------------------------
def make_horizontal_stray_light_array(cellpsf: List[np.ndarray], array_window_specs, y0=1000.5, x0=752.5):
    """
    This is an example that shows how to make the horizontal stray light convolution matrix. Vignetting is set to an
    empty array which is detected by the python interface and a default of 1.0 is used for vignetting
    """

    NOOFx = array_window_specs[0]
    NOOFy = array_window_specs[1]
    locn = np.array([y0, x0, WINDOW_DETECTOR]).reshape([1, 3])                    # set up the (Y,X) coordinates of the center of the PSF, Y then X
    coeffs = np.ones([1, NOOFy, NOOFx])                                                 # Set all the interpolation coefficients to 1.0
    indices = np.zeros([1, NOOFy, NOOFx])                                               # Set all the PSF indices to zero.
    vignetting = np.zeros([0, 0])
    horiz_stray = generate_horizontal_stray_light_matrix(array_window_specs, vignetting, cellpsf, locn, coeffs, indices)
    return horiz_stray


# ------------------------------------------------------------------------------
#           L1C_demo_UV
# ------------------------------------------------------------------------------
def make_uv_in_field_matrix(cellpsf: List[np.ndarray], array_window_specs, y0=1000.5, x0=752.5):
    """
    This is an example that shows how to make the in-field convolution matrix fo rthe L1C product. This only uses 1
    psf array so it creates an interpolation array (of zeros) that indices the first psf. It also makes an array of
    interpolation coefficients for each psf array. The interpolation array is set to 1.0 everywhere.
    """

    Dx = array_window_specs[4]
    Dy = array_window_specs[5]
    locn = np.array([y0, x0, WINDOW_DETECTOR]).reshape([1, 3])
    coeffs = np.ones([1, Dy, Dx])                                               # Set all the interpolation coefficients to 1.0
    indices = np.zeros([1, Dy, Dx])                                             # Set all the PSF indices to zero.
    a_matrix = generate_convolution_matrix(array_window_specs, cellpsf, locn, coeffs, indices)
    return a_matrix


# ------------------------------------------------------------------------------
#           uv_test_analysis
# ------------------------------------------------------------------------------
def uv_test_analysis():
    """
    Runs a test that checks out the deconvolution algorithm. This builds the infield and horizontal stray light
    deconvolution matrices and test data. It then builds the inverse matrices and solves the inversion.
    """
    channel='uv'
    datevalue='2020-10-20T00:00:00.000000'
    wavelen = 550.0

    utc = sktime.ut_to_datetime(datevalue)
    psf_settings = PSFSpecs(channel, datevalue)[wavelen]                                                         # Fetch the PSF wavelength settings for the requested channel and date
    convolution_settings = psf_settings.convolution_matrix_settings                                              # Get the convolution settings from the wavelength entry
    detector_settings = DetectorSpecs( channel, window_config_name=convolution_settings.detector_window)                # Get the detector window setting using the settings described in the PSF settings
    array_window_specs = detector_settings.window_specs                                                                 # load in the window settings
    psf_image = psf_settings[0]

    SOFSx = array_window_specs.SOFSx                                                                                       # X Offset of science window in detector window
    SOFSy = array_window_specs.SOFSy                                                                                    # Y Offset of science window in detector window
    Sx = array_window_specs.Sx                                                                                          # X Size of science window
    Sy = array_window_specs.Sy                                                                                          # Y Size of science window
    LBINx = array_window_specs.LBINx
    LBINy = array_window_specs.LBINy
    KBINx = array_window_specs.KBINx
    KBINy = array_window_specs.KBINy
    Kx = array_window_specs.Kx
    Ky = array_window_specs.Ky

    profile, height = make_uv_test_signal_gauss_feature(Sy)                                                             # Make a 1-D theoretical true profile. Its a simple exponential decay
    true_image = make_true_answer(profile, array_window_specs)                                                          # Make a true image that spans the entire noof window including horizontal stray light
    true_outoffield = np.array(true_image)                                                                              # Make a true image
    true_outoffield[:, SOFSx:SOFSx + Sx] = 0                                                                            # that only has the horizontal out-of-field

                                                                                                                        # Load in the list of 1 PSF arrays
    y0 = psf_image.center_of_psf.y
    x0 = psf_image.center_of_psf.x
    psf = PSFTools.load_psf_from_filename(psf_image.psf_filename)                                                                     # get the psf array from the first element.
    newpsf, y0, x0 = PSFTools.shift_psf_to_image_centre(psf, y0, x0)                                                    # Shift the PSF so it is properly centered (as returned in y0 and x0)
    psf = newpsf

    measured_signal = scipy.signal.fftconvolve(true_image, psf, mode='same')                                            # Calculate the measured signal using a convolution
    measured_signal[0:SOFSy, :] = 0                                                                                     # wipe out the lower
    measured_signal[SOFSy + Sy:, :] = 0                                                                                 # and upper parts of the image (these can be left but are never used)

    true_outoffield_stray = scipy.signal.fftconvolve(true_outoffield, psf, mode='same')                                 # Calculate how much stray light scatters in the science window
    true_outoffield_stray[0:SOFSy, :] = 0
    true_outoffield_stray[SOFSy + Sy:, :] = 0

    a, awindowspecs, acoords = L1CConvolutionMatrix.load_matrix_file(convolution_settings.infield_matrix)
    h, hwindowspecs, hcoords = L1CConvolutionMatrix.load_matrix_file(convolution_settings.horizontal_noof_matrix)
    b = a + h
    Binv = L1CConvolutionMatrix(np.linalg.inv(b.matrix))                                                                              # Get the deconvolution matrix of the corrected in-field stray light by taking the inverse
    Ainv = L1CConvolutionMatrix(np.linalg.inv(a.matrix))

    measured_l1c = make_l1c_from_detector_window_image(measured_signal, array_window_specs)
    true_l1c = make_l1c_from_detector_window_image(true_image, array_window_specs)                                # make the true true answer. This is what we want to retrieve.
    true_stray_l1c = make_l1c_from_detector_window_image(true_outoffield_stray, array_window_specs)                 # make the true stray light scattered

    stray_matrix = h @ true_l1c
    infield_matrix = a @ true_l1c
    answer = Binv @ measured_l1c

    ZBINy = LBINy * KBINy
    ZBINx = LBINx * KBINx
    idx = Kx//2
    h = height.reshape( [Sy//ZBINy,ZBINy])[:,ZBINy//2]
    yt = true_l1c[:, idx]
    ym = measured_l1c[:, idx]
    ya = answer[:, idx]
    yin = infield_matrix[:, idx]
    yoof = stray_matrix[:, idx]
    diff = np.abs((yt - ya) / yt)

    plt.figure(1)
    plt.clf()
    plt.plot(np.log10(yt), h, 'k-', linewidth=0.5, label='True Profile')
    plt.plot(np.log10(ym), h, 'b-', linewidth=0.5, label='Measured Profile')
    plt.plot(np.log10(ya), h, 'r-', linewidth=0.5, label='Retrieved Profile')
    plt.plot(np.log10(yin), h, 'm-', linewidth=0.5, label='Infield signal')
    plt.plot(np.log10(yoof), h, 'g-', linewidth=0.5, label='Horiz Noof signal')
    plt.xlabel('Log10 Signal')
    plt.ylabel('Nominal height km')
    plt.title('Infield + Horiz Stray Light. {} heights'.format(Ky))
    plt.legend()

    plt.figure(2)
    plt.clf()
    plt.plot(diff*100.0, h)
    plt.xlabel('Percentage difference')
    plt.ylabel('Nominal height km')
    plt.title('Inversion Performance. {} heights'.format(Ky))

    plt.show(block=False)
    print('Done')


if __name__ == '__main__':
    uv_test_analysis()
