"""
This is the first draft of a vertical stray light removal tool. This study was initiated in June
2021 in preparation for the ALTIUS PKP#4 in July 2021. The code implements the vertical stray light algorithm developed
in Feb/March 2021 for PKP#3.
"""

from typing import Tuple
import os.path
import numpy as np
import xarray as xr
import netCDF4
import argparse
from altius_l1pp_preflight.configuration import AltiusConfigurationKey, OpticsSpecs, override_yaml_base_folder
from altius_l1pp_preflight.stray_light import AltiusStrayLightAlgorithm


# ------------------------------------------------------------------------------
#           write_results_to_netcdf
# ------------------------------------------------------------------------------
def write_results_to_netcdf(filename: str, true_signal: np.ndarray, true_error: np.ndarray):
    """
    Writes the results to a netcdf file.
    """
    basedir, name = os.path.split(filename)             # make sure the base folder exists.
    os.makedirs(basedir, exist_ok=True)

    if os.path.exists(filename):                        # If the file already exists then delete it
        os.remove(filename)

    matlabdata = true_signal.transpose()                # Place data back into the same format as it came in.
    matlaberror = true_error.transpose()
    datavars = {}
    datavars['l1c_true_signal'] = xr.DataArray(data=matlabdata, name='l1c_true_signal', dims=["Kx", "Ky"])
    datavars['l1c_true_error'] = xr.DataArray(data=matlaberror, name='l1c_true_error', dims=["Kx", "Ky"])
    ds = xr.Dataset(data_vars=datavars)                 # Make the data set
    ds.to_netcdf(filename)                              # and save it to the file.


# ------------------------------------------------------------------------------
#           def read_sps_input_file(filename: str)
# ------------------------------------------------------------------------------
def read_sps_input_file(filename: str) -> Tuple[str, str, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Reads in the input file coming from the Qinetiq SPS system  and returns the signal, error, altitude and angle as
    4 arrays.
    """

    data = netCDF4.Dataset(filename)
    datestr: str = data.date
    channel: str = data.channel
    wavelen: float = float(data.wavelength)
    measured_signal = np.array(data['measured_signal']).transpose()
    measured_error = np.array(data['measured_error']).transpose()
    altitude = np.array(data['altitude'])
    angle = np.array(data['angle'])
    return channel.lower(), datestr, wavelen, measured_signal, measured_error, altitude, angle


# ------------------------------------------------------------------------------
#           main
# ------------------------------------------------------------------------------
def main():
    appname = "remove_vertical_stray"
    parser = argparse.ArgumentParser(description="Altius Vertical Stray Light Removal Tool")
    parser.add_argument('input_filename')
    parser.add_argument('output_filename')
    parser.add_argument('--yaml_folder', action='store', help='The folder of the USASK YAML configuration files', default="")
    commandline = parser.parse_args()

    inputfilename = commandline.input_filename
    outputfilename = commandline.output_filename
    yamlfolder = commandline.yaml_folder

    if (yamlfolder is not None) and (len(yamlfolder) > 0):
        override_yaml_base_folder(yamlfolder)

    channel, datestr, wavelen, measured_signal, measured_error, altitude, angles = read_sps_input_file(inputfilename)
    config = AltiusConfigurationKey(channel=channel, wavelength=wavelen, datestr=datestr)
    algorithm = AltiusStrayLightAlgorithm(config)
    true_estimate = algorithm.estimate_true_profile(measured_signal, measured_error, altitude, angles)
    write_results_to_netcdf(outputfilename, true_estimate, measured_error)


if __name__ == '__main__':
    main()
