import unittest
import os
from altius_l1pp_preflight.configuration import OpticsSpecs, PSFSpecs, SRFUVSpecs, AltiusConfigurationKey, Coords, WindowMappingType
from altius_l1pp_preflight.configuration.load_specs import check_specs_exist_in_user_space
from altius_l1pp_preflight.psf import PSFTools, make_convolution_matrices
from altius_l1pp_preflight.srf import SRFUV
from altius_l1pp_preflight.radiance_simulator import AltiusL1ppSimpleSkimpyInterface, SimulationRadianceDataBISA


# ------------------------------------------------------------------------------
#           AltiusL1ppTestConfiguration
# ------------------------------------------------------------------------------
class AltiusL1ppTestConfiguration(unittest.TestCase):

    def test_config_location(self):
        '''
        Ensure the configuration is properly installed in the user area
        '''
        check_specs_exist_in_user_space()

    def test_skconfig_issue(self):
        envvar = os.environ.get('ALTIUS_L1PP_USASK_DATABASE')                                                       # Do a quick test
        os.environ['ALTIUS_L1PP_USASK_DATABASE'] = r'PQD:/TEST_DIR_VAL'                                             # to verify that a bug in skconfig is no longer an issues
        config = AltiusConfigurationKey(channel='vis', wavelength=550.0, datestr='2020-10-20')
        psfspecs = PSFSpecs(config)[config.wavelength]
        filename = psfspecs.convolution_matrix_settings.infield_matrix_settings['filename']
        self.assertEqual('PQD:/TEST_DIR_VAL/PSF/2020-10-20/psf-conv-matrix_vis_2020-10-20_550.nc', filename, 'You are not reading the configuration properly. You need to ensure package skconfig, version 0.3.3 or higher is installed')
        if envvar is not None:
            os.environ['ALTIUS_L1PP_USASK_DATABASE'] = envvar
        else:
            os.environ.pop('ALTIUS_L1PP_USASK_DATABASE')

    def test_environ_vars_are_set(self):
        envvar = os.environ.get('ALTIUS_L1PP_USASK_DATABASE')
        self.assertIsNotNone(envvar, 'The environment variable <ALTIUS_L1PP_USASK_DATABASE> is not set. This will create problems')

    def test_psf_vis_config(self):
        config = AltiusConfigurationKey(channel='vis', wavelength=550.0, datestr='2020-10-20')                       # this test requires the
        image = PSFTools.load_psf(config, 0)                                                                         # load in the image, it should return a 2 element tuple
        self.assertTrue(isinstance(image, tuple))
        s = image[0].shape
        self.assertEqual(2000, s[0])
        self.assertEqual(1504, s[1])

    def test_srf_uv_config(self):
        config = AltiusConfigurationKey(channel='uv', datestr='2020-08-18')
        srfspecs = SRFUVSpecs(config)
        srfuv = SRFUV(config)
        srfuv.select_SRF(300.0)
        settings, w0 = srfuv.airgap_settings(300.0)
        y = srfuv.fpi_srf

    def test_detector_config(self):
        config = AltiusConfigurationKey(channel='vis', wavelength=550.0, datestr='2020-10-20')
        dspecs = OpticsSpecs(config)

    def test_mapping(self):
        config = AltiusConfigurationKey(channel='vis', wavelength=550.0, datestr='2020-10-20')
        optics = OpticsSpecs(config)
        location = Coords(x=8.2, y=251.2, windowtype=WindowMappingType.WINDOW_L1C)
        detector = optics.detector_specs.window_hierarchy.convert_to(location, WindowMappingType.WINDOW_DETECTOR)
        newlocation = optics.detector_specs.window_hierarchy.convert_to(detector, WindowMappingType.WINDOW_L1C)
        self.assertAlmostEqual(location.x, newlocation.x)
        self.assertAlmostEqual(location.y, newlocation.y)
        self.assertAlmostEqual(detector.x, 1230.0)
        self.assertAlmostEqual(detector.y, 1003.6)
        radians = optics.pixel_to_radians(location)

    def test_simple_skimpy_interface(self):
        configkey = AltiusConfigurationKey('vis')
        optics = OpticsSpecs(configkey)
        simulator = AltiusL1ppSimpleSkimpyInterface(optics, tp_latitude=38.0)
        radiance, latitude, longitude, altitude = simulator.radiance(550.0, albedo=1.0)
        self.assertAlmostEqual(latitude[750], 38.000077042437994, places=5)
        self.assertAlmostEqual(altitude[750], 50066.414668127894, places=5)
        self.assertAlmostEqual(radiance[750] / 541922109590.63934, 1.0, places=6)

    def test_bisa_stimuli(self):
        configkey = AltiusConfigurationKey('vis')
        optics = OpticsSpecs(configkey)
        stimuli = SimulationRadianceDataBISA()
        signal, error = stimuli.make_noof_image(optics, 3, 4)
        self.assertAlmostEqual(signal[3000, 3000], 6010.094451515401, places=5)
        self.assertAlmostEqual(signal[1000, 0], 27883.177782753806, places=5)

    def test_make_convolution(self):
        make_convolution_matrices.make_infield_convolution_matrix('uv', '2020-10-20', 300.0, filename='psf300.nc', overwrite=True, debug_mode=True)


if __name__ == "__main__":
    tests = AltiusL1ppTestConfiguration()
    unittest.main()
